// flare.C

/***********************************************************************/
/*                                                                     */
/* THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,  */
/* AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH  */
/* (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL */
/* WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF  */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY       */
/* WARRANTIES OR NON INFRINGEMENT. THE USER ASSUMES ALL LIABILITY AND  */
/* RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR  */
/* NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE    */
/* FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE. Without limiting    */
/* the generality of the foregoing, neither the author, nor JAIR, nor  */
/* JAIR's publishers and distributors, warrant that the Source Code    */
/* will be error-free, will operate without interruption, or will meet */
/* the needs of the user.                                              */
/*                                                                     */
/***********************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <sgtty.h>
#include <strings.h>
#include <sys/times.h>
#include "header.H"
#include "vec_s.H"
#include "network.H"
#include "flare.H"

#define BUFSIZE 40
#define NUM_TRIALS 10 
#define NUM_WAYS 10 
#define BATCH .5
#define INTERACTIVE 1
#define W_PREC 1
#define G_PREC 2
#define WO_PREC 0

#define CLK_TCK 60

Header *info;
char   header_file[40], tr_file[40], te_file[40], pr_file[40];
struct sgttyb old_env, new_env;

void main(int argc, char *argv[])
{
  int prec_flag;

  ioctl(fileno(stdin), TIOCGETP, &old_env);
  new_env = old_env;
  new_env.sg_flags |= CBREAK;
  ioctl(fileno(stdin), TIOCSETP, &new_env);

  strcpy(header_file, argv[2]);
  strcpy(tr_file, argv[2]);
  strcpy(te_file, argv[2]);
  strcpy(pr_file, argv[2]);
  prec_flag = WO_PREC;

  if (argc == 4)
  {
    strcat(pr_file, ".pr");
    if (!strcmp(argv[3], "g"))
      prec_flag = G_PREC;
    else
      prec_flag = W_PREC;
  }
  strcat(header_file, ".hdr");
  info = new Header(header_file);
  strcat(tr_file, ".raw");
  if (!strcmp(argv[1], "b"))
    batch(prec_flag);
  else if (!strcmp(argv[1], "i"))
  {
    strcat(te_file, ".tst");
    interactive(prec_flag);
  }
  else
    cross_val(prec_flag);
  delete info;

  ioctl(fileno(stdin), TIOCSETP, &old_env);
}

void batch(int p_flag)
{
  int         i, avg_n;
  float       gen, avg;
  Network    *Net;
  Vector_set *TrSet, *PrSet;
  clock_t tm, start_time, end_time;
  struct tms Time;

  avg = 0.0; avg_n = 0; tm = 0;
  for (i = 0; i < NUM_TRIALS; i++)
  {
    Net = new Network();
    if (p_flag == W_PREC)
    { 
      PrSet = new Vector_set(pr_file, PREC);
      add_precepts(Net, PrSet);
    }
    else if (p_flag == G_PREC)
      generate_precepts(Net, pr_file);
    TrSet = new Vector_set(tr_file, TRAIN);
    times(&Time);
    start_time = Time.tms_utime;
    learn(Net, TrSet, (float)info->num_tr_instances * BATCH);
    times(&Time);
    end_time = Time.tms_utime;
    tm += (end_time - start_time);
    gen = generalize(Net, TrSet);
    printf("\ntrial #%d: %f", i+1, gen);
    avg += gen;
    avg_n += Net->size();
    delete Net;
    delete TrSet;
    if (p_flag == W_PREC)
      delete PrSet;
  }
  printf("\n        Generalization:     %f  %d  %d %6.2f\n",
          avg/(float)NUM_TRIALS, (int)((float)info->num_tr_instances*BATCH),
          avg_n/NUM_TRIALS, (double)tm/NUM_TRIALS/CLK_TCK);
}

void interactive(int p_flag)
{
  float      gen;
  Network    *Net;
  Vector_set *TrSet, *TeSet, *PrSet;

  TrSet = new Vector_set(tr_file, TRAIN);
  TeSet = new Vector_set(te_file, TEST);
  Net = new Network();
  if (p_flag == W_PREC)
  {
    PrSet = new Vector_set(pr_file, PREC);
    add_precepts(Net, PrSet);
  }
  else if (p_flag == G_PREC)
    generate_precepts(Net, pr_file);
  learn(Net, TrSet, (float)info->num_tr_instances * INTERACTIVE);
  Net->print();
  gen = generalize(Net, TeSet);
  printf("\n        Generalization:     %f  %d  %d  %d\n", gen,
         info->num_tr_instances, Net->size(), info->num_te_instances);
  delete TrSet;
  delete TeSet;
  if (p_flag == W_PREC)
    delete PrSet;
  delete Net;
}

void cross_val(int p_flag)
{
  int        i, j, avg_n, avg_ex, num_tr, num_te;
  float      gen, avg;
  Network    *Net;
  Vector     *v;
  Vector_set *DataSet, *TrSet, *TeSet, *PrSet;

  DataSet = new Vector_set(tr_file, TRAIN);
  avg = 0.0; avg_n = avg_ex = 0;
  for (i = 1 ; i <= NUM_WAYS; i++)
  {

// construct test set and training set for cross-validation

    num_tr = num_te = 0;
    TeSet = new Vector_set();
    for (j = 0; j < anint(((float)(NUM_WAYS - (i - 1)) / (float)NUM_WAYS) *
                          (float)info->num_tr_instances) - anint(((float)
                          (NUM_WAYS - i) / (float)NUM_WAYS) * (float)
                          info->num_tr_instances); j++)
    {
      num_te++;
      TeSet->add_vector(DataSet->get_next_te_vector_cv());
    }
    TrSet = new Vector_set();
    DataSet->init();
    while ((v = DataSet->get_next_tr_vector_cv()) != NULL)
    {
      num_tr++;
      TrSet->add_vector(v);
    }

// average over 10 trials with random learning order

    int sub_avg_n = 0;
    float sub_avg = 0.0;
    for (j = 0; j < NUM_TRIALS; j++)
    {
      Net = new Network();
      if (p_flag == W_PREC)
      {
        PrSet = new Vector_set(pr_file, PREC);
        add_precepts(Net, PrSet);
      }
      else if (p_flag == G_PREC)
        generate_precepts(Net, pr_file);
      TrSet->reset_cv();
      for (int k = 0; k < num_tr; k++)
        Net->broadcast_NV(TrSet->get_random_vector_cv()); 
      TeSet->init();
      int cnt = 0, dummy1;
      float s = 0.0, dummy;
      while ((v = TeSet->get_next_vector()) != NULL)
      {
        cnt++;
        if (v->output() == Net->execute_UV(v, &dummy, REASONING, &dummy1))
          s++;
      }
      s /= (float)cnt;
      sub_avg += (s * 100.0);
      sub_avg_n += Net->size();
      if (p_flag == W_PREC)
        delete PrSet;
      delete Net;
    }
    avg_ex += num_tr;
    avg += (sub_avg / (float)NUM_TRIALS);
    avg_n += (sub_avg_n / NUM_TRIALS);
    delete TrSet;
    delete TeSet;
  }
  printf("\n        Generalization:     %f  %d  %d \n",
          avg/(float)NUM_WAYS, avg_ex/NUM_WAYS,
          avg_n/NUM_WAYS);
  delete DataSet;
}

void learn(Network *Net, Vector_set *VS, float num_inst) 
{
  int    i;
  Vector *V;

  for (i = 0; i < num_inst; i++)
  {
    V = VS->get_next_test_vector();
    Net->broadcast_NV(V);
  }
}

float generalize(Network *Net, Vector_set *VS)
{
  int    cnt, dummy1;
  float  s, dummy, target_output, computed_output;
  Vector *UV;

  cnt = 0; s = 0.0;
  while ((UV = VS->get_next_test_vector()) != NULL)
  {
    cnt++;
    target_output = UV->output();
    computed_output = Net->execute_UV(UV, &dummy, REASONING, &dummy1);
    if (computed_output == target_output)
      s++;
    delete UV;
  }
  s /= (float)cnt;
  return (s*100.0);
}

void add_precepts(Network *Net, Vector_set *VS)
{
  Vector *NV;
 
  while ((NV = VS->get_next_test_vector()) != NULL)
    Net->broadcast_NV(NV);
}

void generate_precepts(Network *Net, char *fname)
{
  int    i, dummy1;
  char   tmp1[40], tmp2[40];
  float  dummy, result;
  Vector *NV;

  strcpy(tmp1, fname);
  strcpy(tmp2, fname);
  strcat(tmp1, ".cs");
  strcat(tmp2, ".pr");
  Network *tmp_net = new Network();
  Vector_set *tmp_vs = new Vector_set(tmp1, CSRULES);
  add_precepts(tmp_net, tmp_vs);
  Vector_set *tmp_ts = new Vector_set(tmp2, PREC);
  
  while ((NV = tmp_ts->get_next_test_vector()) != NULL)
  {
    result = tmp_net->execute_UV(NV, &dummy, REASONING, &dummy1);
    if (result != DK)
    {
      NV->set_attribute(NV->target, result);
      for (i = NV->target + 1; i < info->num_attributes; i++)
        NV->set_attribute(i, (float)DC);
      Net->broadcast_NV(NV);
    }
  }
  delete tmp_vs;
  delete tmp_ts;
  delete tmp_net;
}
