// network.C

/***********************************************************************/
/*                                                                     */
/* THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,  */
/* AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH  */
/* (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL */
/* WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF  */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY       */
/* WARRANTIES OR NON INFRINGEMENT. THE USER ASSUMES ALL LIABILITY AND  */
/* RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR  */
/* NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE    */
/* FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE. Without limiting    */
/* the generality of the foregoing, neither the author, nor JAIR, nor  */
/* JAIR's publishers and distributors, warrant that the Source Code    */
/* will be error-free, will operate without interruption, or will meet */
/* the needs of the user.                                              */
/*                                                                     */
/***********************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "header.H"
#include "vec_s.H"
#include "node.H"
#include "network.H"

extern Header *info;

Network::Network()
{
  head = current = NULL;
  nb_nodes = 0;
}

Network::~Network()
{
  current = head;
  while (nb_nodes != 0)
    delete_current_node();
}

int Network::size()
{
  return (nb_nodes);
}

void Network::print()
{
  static int i;
  Elementn   *elt;

  printf("\n\n  Network: \n");
  elt = head;
  for (i = 0; i < nb_nodes; i++)
  {
    elt->nd->print();
    elt = elt->next;
  }
  printf("\n");
}

void Network::add_node(Node *n)
{
  Elementn *elt = new Elementn;

  elt->nd = n;
  if (nb_nodes == 0)
    elt->next = NULL;
  else
    elt->next = head;
  head = current = elt;
  nb_nodes++;
}

Elementn *Network::delete_current_node()
{
  Elementn *elt;

  if (current == head) {
    head = current->next;
    delete current->nd;
    delete current;
    current = head;
  }
  else {
    elt = head;
    while (elt->next != current)
      elt = elt->next;
    elt->next = current->next;
    delete current->nd;
    delete current;
    current = elt->next;
  }
  nb_nodes--;
  return (current);
}

void Network::broadcast_NV(Vector *v)
{
  int      match, concord, diff, n_match, cf, self_delete_flag;
  Elementn *win;
  float    out, winact;

  Node *n = new Node(v);
  self_delete_flag = OFF;

  cf = 1;
  out = execute_UV(v, &winact, LEARNING, &cf);
  if (!cf) return;

  win = current;
  if (!win)
  {
    add_node(n);
    return;
  }
  match = win->nd->vector()->matching_state(v);
  concord = win->nd->vector()->concordant(v);
  diff = abs(win->nd->vector()->num_not_dc - v->num_not_dc);
  n_match = win->nd->vector()->near_match(v);
  if (match == EQUAL)
  {
    win->nd->inc_counters((int)v->output());
    win->nd->vector()->set_output(win->nd->max_counter());
    delete n;
  }
  else if (match == SUBSET && concord)
  {
    win->nd->num_covers++;
    delete n;
  }
  else if (match == DISCRIMINATED && concord && n_match && diff <= 1)
  {
    if (win->nd->vector()->num_not_dc <= v->num_not_dc &&
        win->nd->vector()->num_not_dc > 1)
    {
      Vector *new_vec = new Vector(win->nd->vector());
      new_vec->drop_variable(v);
      if (win->nd->vector()->pty > v->pty)
        new_vec->pty = win->nd->vector()->pty;
      else
        new_vec->pty = v->pty;
      Node *nn = new Node(new_vec);
      nn->num_covers = win->nd->num_covers;
      add_node(nn);
      self_delete_flag = ON;
      delete n;
    }
    else if (v->num_not_dc > 1)
    {
      Vector *new_vec = new Vector(n->vector());
      new_vec->drop_variable(win->nd->vector());
      if (win->nd->vector()->pty > v->pty)
        new_vec->pty = win->nd->vector()->pty;
      else
        new_vec->pty = v->pty;
      Node *nn = new Node(new_vec);
      nn->num_covers = win->nd->num_covers;
      add_node(nn);
      self_delete_flag = ON;
      delete n;
    }
    else
      add_node(n);
  }
  else if (match == SUPERSET && concord)
  {
    n->num_covers++;
    self_delete_flag = ON;
    add_node(n);
  }
  else add_node(n);

  if (self_delete_flag)
  {
    current = win;
    delete_current_node();
  }
}

float Network::attempt_to_assert(Vector *V, int i, float *distance,
                                 float target_out, int flag, int *cf)
{
  int      found;
  float    maxact, curact;
  Elementn *maxelt, *curelt;

  maxelt = curelt = head;
  maxact = 10000000.0;
  found = 0;
  while (curelt)
  {
    if (curelt->nd->vector()->target == i)
    {
      found = 1;
      curact = curelt->nd->activation(V);
      int closer = (curact < maxact);
      int equidistant = (curact == maxact);
      int more_specific = (curelt->nd->vector()->num_not_dc >
                           maxelt->nd->vector()->num_not_dc);
      int same_spec = (curelt->nd->vector()->num_not_dc ==
                       maxelt->nd->vector()->num_not_dc);
      int greater_pty = (curelt->nd->vector()->pty >
                         maxelt->nd->vector()->pty);
      int same_pty = (curelt->nd->vector()->pty ==
                      maxelt->nd->vector()->pty);
      int conf_default = ((maxelt->nd->vector()->matching_state(curelt->
                           nd->vector()) == OVERLAP) && !maxelt->nd->
                           vector()->concordant(curelt->nd->vector()) &&
                           curact == 0.0);   // distance must be 0 for defaults
      int greater_cover = (curelt->nd->num_covers > maxelt->nd->num_covers);
      if (closer)
      {
        maxact = curact;
        maxelt = curelt;
      }
      else if (equidistant && more_specific)
      {
        maxact = curact;
        maxelt = curelt;
      }
      else if (equidistant && same_spec && greater_pty)
      {
        maxact = curact;
        maxelt = curelt;
      }
      else if (equidistant && same_spec && same_pty)
      {
        if (conf_default && flag == LEARNING)
        {
          if (maxelt->nd->vector()->output() == target_out)
          {
            maxelt->nd->dyn_pty++;
            *cf = 0;		// if cf=0 the vector should not be added
          }
          else if (curelt->nd->vector()->output() == target_out)
          {
            curelt->nd->dyn_pty++;
            *cf = 0;
          }
        }
        int greater_dyn_pty = (curelt->nd->dyn_pty > maxelt->nd->dyn_pty);
        int same_dyn_pty = (curelt->nd->dyn_pty == maxelt->nd->dyn_pty);
        if (greater_dyn_pty || (same_dyn_pty && greater_cover))
        {
          maxact = curact;
          maxelt = curelt;
        }
      }
    }
    curelt = curelt->next;
  }
  *distance = maxact;
  if (found)
  {
    current = maxelt;
    return (maxelt->nd->vector()->output());
  }
  else
  {
    current = NULL;
    return ((float)DK);
  }
}

float Network::execute_UV(Vector *UV, float *winact, int flag, int *cf)
{
  int      i, done;
  float    predicted, target_out;
  Elementn *curelt;

// save actual target value of output, reset to ? for reasoning
  target_out = UV->output();
  UV->set_attribute(UV->target, DK);

// create a copy of UV
  Vector *tmp1 = new Vector(UV);

// completion - uses first rule found 
  for (i = 0; i < info->num_attributes; i++)
    if (tmp1->asserted(i) && tmp1->target != i)
    {
      curelt = head;
      while (curelt && (curelt->nd->vector()->target != i ||
                        !curelt->nd->vector()->same_attr_val(UV, i) ||
                        !curelt->nd->vector()->def))
        curelt=curelt->next;
      if (curelt)
        tmp1->complete(curelt->nd->vector());
    }

// if completion set the output, copy into UV, reset actual output, done
  if (tmp1->output() != DK)
  {
    UV->copy(tmp1);
    UV->set_attribute(UV->target, target_out);
    return (tmp1->output());
  }

// assert all possible unknown attributes other than target
  float threshold = 0.0;
  done = 0;
  while (!done)
  {
    done = 1;
    for (i = 0; i < info->num_attributes; i++)
    {
      if (!tmp1->asserted(i)/*dont_know(i)*/ && UV->target != i)
      {
        predicted = attempt_to_assert(tmp1, i, winact, target_out, flag, cf);
        if (predicted != DK && *winact <= threshold)
        {
          done = 0;
          UV->set_attribute(i, predicted);
        }
      }
    }
    tmp1->copy(UV);
  }

// predict the value of the target
  predicted = attempt_to_assert(UV, UV->target, winact, target_out, flag, cf);

// reset actual output value, return predicted value (can be asserted elsewhere)
  UV->set_attribute(UV->target, target_out);
  return (predicted);
}
