// network.H

/***********************************************************************/
/*                                                                     */
/* THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,  */
/* AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH  */
/* (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL */
/* WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF  */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY       */
/* WARRANTIES OR NON INFRINGEMENT. THE USER ASSUMES ALL LIABILITY AND  */
/* RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR  */
/* NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE    */
/* FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE. Without limiting    */
/* the generality of the foregoing, neither the author, nor JAIR, nor  */
/* JAIR's publishers and distributors, warrant that the Source Code    */
/* will be error-free, will operate without interruption, or will meet */
/* the needs of the user.                                              */
/*                                                                     */
/***********************************************************************/


// This class defines a network as a list of nodes

#if !defined(NETWORK_H)
#define NETWORK_H

#define LEARNING  1
#define REASONING 2

#include "vec_s.H"
#include "node.H"

    typedef struct eln
    {
      Node *nd;
      eln  *next;
    } Elementn;                  // list element - network = list of nodes

class Network
{
  public:
    Network();                                    // constructor
    int size();                                   // returns size of network
    void print();  		                  // prints network info
    void broadcast_NV(Vector *v);                 // broadcasts v to network
    float execute_UV(Vector *v, float *f, int f, int *cf);
            // executes network with v
    float attempt_to_assert(Vector *V, int i, float *f,
                            float target_out, int flag, int *cf);
            // attempts to assert ith attr. of V, where V has target value
            // target_out, f is the maxact, flag is for confl. defaults in
            // learning, and cf tells whether except. should be stored
    ~Network();                                   // destructor
  private:
    Elementn    *head, *current;         // head and current ptrs into list
    int         nb_nodes;                // nb of nodes in network
    Elementn    *delete_current_node();  // deletes current node, returns successor
    void        add_node(Node *n);       // adds node for vector v
};

#endif

