// node.C

/***********************************************************************/
/*                                                                     */
/* THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,  */
/* AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH  */
/* (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL */
/* WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF  */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY       */
/* WARRANTIES OR NON INFRINGEMENT. THE USER ASSUMES ALL LIABILITY AND  */
/* RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR  */
/* NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE    */
/* FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE. Without limiting    */
/* the generality of the foregoing, neither the author, nor JAIR, nor  */
/* JAIR's publishers and distributors, warrant that the Source Code    */
/* will be error-free, will operate without interruption, or will meet */
/* the needs of the user.                                              */
/*                                                                     */
/***********************************************************************/


#include <stdio.h>
#include "header.H"
#include "vector.H"
#include "node.H"

extern Header *info;

Node::Node(Vector *v)
{
  int i;

  vect = v;
  for (i = 0; i < info->num_states[v->target]; i++)
    if (i == (int)v->output())
      counters[i] = 1;
    else
      counters[i] = 0;
  num_covers = dyn_pty = 0;
}

void Node::inc_counters(int ind)
{
  counters[ind]++;
}

int Node::counter(int ind)
{
  return (counters[ind]);
}

void Node::copy_counters(Node *n)
{
  int i;

  for (i = 0; i < info->num_states[vect->target]; i++)
    counters[i] = n->counters[i];
  num_covers = n->num_covers;
}

int Node::max_counter()
{
  int i, max, new_out;

  max = -1;
  for (i = 0; i < info->num_states[vect->target]; i++)
    if (counters[i] > max)
    {
      max = counters[i];
      new_out = i;
    }
  return (new_out);
}

float Node::activation(Vector *v)
{
  float act;

  if (vect->num_not_dc == 0) return (0.0);
  act = vect->distance(v)/(float)vect->num_not_dc;
  return (act);
}

Vector *Node::vector()
{
  return (vect);
}

void Node::print()
{
  vect->print();
  printf("  cover=%d dyn_pty=%d", num_covers, dyn_pty);
}

Node::~Node()
{
}

