// vec_s.C

/***********************************************************************/
/*                                                                     */
/* THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,  */
/* AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH  */
/* (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL */
/* WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF  */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY       */
/* WARRANTIES OR NON INFRINGEMENT. THE USER ASSUMES ALL LIABILITY AND  */
/* RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR  */
/* NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE    */
/* FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE. Without limiting    */
/* the generality of the foregoing, neither the author, nor JAIR, nor  */
/* JAIR's publishers and distributors, warrant that the Source Code    */
/* will be error-free, will operate without interruption, or will meet */
/* the needs of the user.                                              */
/*                                                                     */
/***********************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include "header.H"
#include "vec_s.H"

#define MAX_ATTRIBUTES 100
#define L 1
#define T 2

extern Header *info;

Vector_set::Vector_set()
{
  head = current = NULL;
  num_vectors = 0;
}

void Vector_set::add_vector(Vector *v)
{
  Elementv *elt = new Elementv;

  elt->vect = v;
  elt->tag = FREE;
  elt->cvtag = FREE;
  elt->next = NULL;
  if (!head)
    head = elt;
  else
    current->next = elt;
  current = elt;
  num_vectors++;
}

Vector_set::Vector_set(char *filename, int flag)
{
  static int i, j, len, target, pty, d;
  static float vec_in[MAX_ATTRIBUTES];
  static char s[10];

  if ((vecf = fopen(filename, "r")) == NULL)
  {
    printf("\n  ERROR: cannot open data set file %s  (vec_s.C)....\n", filename);
    exit(1);
  }
  head = current = NULL;
  if (flag == TRAIN)
    len = info->num_tr_instances;
  else if (flag == TEST)
    len = info->num_te_instances;
  else if (flag == PREC)
    len = info->num_precepts;
  else
    len = info->num_csrules;
  num_vectors = len;
  for (i = 0; i < len; i++)
  {
    for (j = 0; j < info->num_attributes; j++)
    {
      fscanf(vecf, "%s", s);
      if (s[0] == '*')
        vec_in[j] = (float)DC;
      else if (s[0] == '?')
        vec_in[j] = (float)DK;
      else
        vec_in[j] = atof(s);
    }
    fscanf(vecf, "%s", s);
    target = atoi(s) - 1;
    fscanf(vecf, "%s", s);
    pty = atoi(s);
    fscanf(vecf, "%s", s);
    d = atoi(s);
    Vector *V = new Vector(vec_in, target, pty, d);
    Elementv *elt = new Elementv;
    elt->vect = V;
    elt->tag = FREE;
    elt->cvtag = FREE;
    if (i == 0)
    {
      elt->next = NULL;
      head = elt;
    }
    else
      current->next = elt;
    current = elt;
  }
  current->next = NULL;
  fclose(vecf);
}

Vector_set::~Vector_set()
{
  Elementv *elt;

  current = head;
  while (current)
  {
    elt = current->next;
    delete current;
    current = elt;
  }
  num_vectors = 0;
}

Vector *Vector_set::get_next_training_vector()
{ 
  long int i, sel;

  do
  {
    sel = random() % num_vectors; 
    current = head;
    for (i = 0; i < sel; i++)
      current = current->next;
  }
  while (current->tag);
  current->tag = MARKED;
  return (current->vect);
}

Vector *Vector_set::get_next_test_vector()
{
  current = head;
  while (current && current->tag)
    current = current->next;
  if (current)
  {
    current->tag = MARKED;
    return (current->vect);
  }
  else
    return (NULL);
}

Vector *Vector_set::get_next_tr_vector_cv()
{
  Elementv *tmp;

  while (current && current->tag == T)
  {
    current->tag = L;
    current = current->next;
  }
  if (current)
  {
    tmp = current;
    current = current->next;
    return (tmp->vect);
  }
  else
    return (NULL);
}

void Vector_set::init()
{
  current = head;
}

void Vector_set::reset_cv()
{
  current = head;
  while (current)
  {
    current->cvtag = FREE;
    current = current->next;
  }
}

Vector *Vector_set::get_random_vector_cv()
{
  long int i, sel;

  do
  {
    sel = random() % num_vectors;
    current = head;
    for (i = 0; i < sel; i++)
      current = current->next;
  }
  while (current->cvtag);
  current->cvtag = MARKED;
  return (current->vect);
}

Vector *Vector_set::get_next_te_vector_cv()
{
  long int i, sel;

  do
  {
    sel = random() % num_vectors;
    current = head;
    for (i = 0; i < sel; i++)
      current = current->next;
  }
  while (current->tag);
  current->tag = T;
  return (current->vect);
}

Vector *Vector_set::get_next_vector()
{
  Elementv *tmp;

  tmp = current;
  if (tmp)
  {
    current = current->next;
    return (tmp->vect);
  }
  else
    return (NULL);
}




