// vector.C

/***********************************************************************/
/*                                                                     */
/* THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,  */
/* AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH  */
/* (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL */
/* WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF  */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY       */
/* WARRANTIES OR NON INFRINGEMENT. THE USER ASSUMES ALL LIABILITY AND  */
/* RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR  */
/* NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE    */
/* FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE. Without limiting    */
/* the generality of the foregoing, neither the author, nor JAIR, nor  */
/* JAIR's publishers and distributors, warrant that the Source Code    */
/* will be error-free, will operate without interruption, or will meet */
/* the needs of the user.                                              */
/*                                                                     */
/***********************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "header.H"
#include "vector.H"

#define CLOSE_LIN .2

#define ABS(x) ((x)>0 ? (x) : (-(x)))

extern Header *info;

Vector::Vector(Vector *other)
{
  static int i;

  vector = new float[info->num_attributes];
  for (i = 0; i < info->num_attributes; i++)
    vector[i] = other->vector[i];
  target = other->target;
  pty = other->pty;
  num_not_dc = other->num_not_dc;
  def = other->def;
}

Vector::Vector(float v[], int t, int p, int d)
{
  static int i;

  vector = new float[info->num_attributes];
  target = t;
  pty = p;
  num_not_dc = 0;
  def = d;
  for (i = 0; i < info->num_attributes; i++)
  {
    vector[i] = v[i];
    num_not_dc += (!dont_care(i) && i != target);
  }
}

void Vector::copy(Vector *other)
{
  static int i;

  for (i = 0; i < info->num_attributes; i++)
    vector[i] = other->vector[i];
}

int Vector::same_attr_val(Vector *other, int i)
{
  return (vector[i] == other->vector[i]);
}

void Vector::complete(Vector *other)
{
  static int i;

  for (i = 0; i < info->num_attributes; i++)
    if (other->asserted(i) && other->target != i && dont_know(i))
      vector[i] = other->vector[i];
}

int Vector::same_target(Vector *other)
{
  return (target == other->target);
}

float Vector::output()
{
  return (vector[target]);
}

void Vector::set_output(int new_out)
{
  vector[target] = (float)new_out;
}

void Vector::set_attribute(int ind, float val)
{
  vector[ind] = val;
  if (val == (float)DC)
    num_not_dc--;
}

int Vector::concordant(Vector *other)
{
  return (vector[target] == other->vector[target] && same_target(other));
}

int Vector::dont_care(int i)
{
  return ((vector[i] == (float)DC));
}

int Vector::dont_know(int i)
{
  return ((vector[i] == (float)DK));
}

int Vector::asserted(int i)
{
  return ((!dont_care(i) && !dont_know(i)));
}

void Vector::print()
{
  static int i;

  printf("\n");
  for (i = 0; i < info->num_attributes; i++)
  {
    if (vector[i] == DC)
      printf("* ");
    else if (vector[i] == DK)
      printf("? ");
    else
      printf("%f ", vector[i]);
  }
  printf("  target: %d  pty: %d", target+1, pty);
}

int Vector::near_match(Vector *other)
{
  return (((num_not_dc - matching_cnt(other)) == 1 ||
          (other->num_not_dc - other->matching_cnt(this)) == 1));
}

void Vector::drop_variable(Vector *other)    // drop only if nominal
{
  static int i;

  for (i = 0; i < info->num_attributes; i++)
    if (vector[i] != other->vector[i] && !dont_care(i) &&
        !other->dont_care(i) && info->attr_type[i] == NOMINAL)
    {
      vector[i] = (float)DC;
      num_not_dc--;
    }
}

float Vector::distance(Vector *other)
{
  static int i;
  static float d;

  d = 0.0;
  for (i = 0; i < info->num_attributes; i++)
  {
    if (i != target)
    {
      int dk_ass = (dont_know(i) && other->asserted(i));
      int dk_dk = (dont_know(i) && other->dont_know(i));
      int dk_dc = (dont_know(i) && other->dont_care(i));
      int ass_ass = (asserted(i) && other->asserted(i));
      int ass_dk = (asserted(i) && other->dont_know(i));
      int ass_dc = (asserted(i) && other->dont_care(i));
      if (dk_dc || ass_dc || ass_dk) d += 0.5;
      if (dk_ass || dk_dk) d += 0.5;
      if (ass_ass)
        if (info->attr_type[i] == LINEAR && info->num_states[i] != 0.0)
          d += (float)ABS(vector[i] - other->vector[i]) /
                (float)(info->num_states[i]);
        else if (info->attr_type[i] == NOMINAL)
          d += (float)(vector[i] != other->vector[i]);
    }
  }
  return (d);
}

int Vector::matching_cnt(Vector *other)
{
  static int i, s;

  s = 0;
  for (i = 0; i < info->num_attributes; i++)
    if (vector[i] == other->vector[i]  && !dont_care(i) && i != target)
      s++;
  return (s);
}

int eq(float x, float y, int i)
{
  return (((x == y && info->attr_type[i] == NOMINAL) ||
          (ABS(x - y) < CLOSE_LIN*info->num_states[i] &&
           info->attr_type[i] == LINEAR)));
}

int Vector::matching_state(Vector *other)
{
  static int i, j;

  i = 0;
  while (i == target || (eq(vector[i], other->vector[i], i) &&
         i < info->num_attributes))
    i++;
  if (i == info->num_attributes)
    return (EQUAL);
  j = i;
  while (i == target || ((dont_care(i) || eq(vector[i], other->vector[i], i)) &&
         i < info->num_attributes))
    i++;
  if (i == info->num_attributes)
    return (SUBSET);
  i = j;
  while (i == target || ((other->dont_care(i) || eq(vector[i], other->vector[i],
         i)) && i < info->num_attributes))
    i++;
  if (i == info->num_attributes)
    return (SUPERSET);
  while (j == target || ((other->dont_care(j) || dont_care(j) || eq(vector[j],
         other->vector[j], j)) && j < info->num_attributes))
    j++;
  if (j == info->num_attributes)
    return (OVERLAP);
  return (DISCRIMINATED);
}

Vector::~Vector()
{
  delete vector;
}
