// vector.H

/***********************************************************************/
/*                                                                     */
/* THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,  */
/* AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH  */
/* (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL */
/* WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF  */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY       */
/* WARRANTIES OR NON INFRINGEMENT. THE USER ASSUMES ALL LIABILITY AND  */
/* RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR  */
/* NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE    */
/* FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE. Without limiting    */
/* the generality of the foregoing, neither the author, nor JAIR, nor  */
/* JAIR's publishers and distributors, warrant that the Source Code    */
/* will be error-free, will operate without interruption, or will meet */
/* the needs of the user.                                              */
/*                                                                     */
/***********************************************************************/


// This class defines an Vector as an array of real values with a 
// target attribute and a priority.

#include <math.h>
#include <limits.h>

#if !defined(VECTOR_H)
#define VECTOR_H

#define YES           1 
#define NO            0
#define DC           -75489 
#define DK           -89706
#define EQUAL         1
#define SUBSET        2
#define SUPERSET      3
#define OVERLAP       4
#define DISCRIMINATED 5

class Vector
{
  public:
    Vector(Vector *other);                   // constructor from copy
    Vector(float v[], int t, int p, int d);  // constructor from values
    float output();                          // returns output value
    void set_output(int new_out);            // sets output value to new_out
    void set_attribute(int ind, float val);  // sets indth attr. value to val
    int concordant(Vector *other);           // returns true if self and
                                             // other have same output value
    void print();                            // prints out vector info
    int near_match(Vector *other);           // is vector 1 away from other (0/1)
    void drop_variable(Vector *other);       // drop var in vector that is 1
                                             // away from other, only if var is
                                             // nominal
    float distance(Vector *other);           // returns dist(self, other)
    int matching_state(Vector *other);       // returns match state between this
                                             // and other
    ~Vector();                               // destructor
    void copy(Vector *other);
    void complete(Vector *otehr);
    int asserted(int i);
    int dont_know(int i);
    int same_attr_val(Vector *other, int i); // true if self and other have same
                                             // ith attribute value
    int same_target(Vector *other);          // returns true if self and other
                                             // have same target
    int num_not_dc, target, pty, def;
  private:
    float *vector;                           // the vector itself
    int matching_cnt(Vector *other);
    int dont_care(int i);
};

#endif

